import argparse
import logging
import os

from webdriver_recorder.models import Report
from webdriver_recorder.report_exporter import ReportExporter

here = os.path.dirname(os.path.abspath(__file__))


def get_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        "Regenerates a report from an existing report.json using a custom or default template. "
        "This makes it easier to tweak the template without having to run a full test suite each time, "
        "as long as the report.json appears correct."
    )
    parser.add_argument("--input-json", "-i", required=True, help="The path to the report.json you want to load.")
    parser.add_argument(
        "--output-dir",
        "-o",
        required=False,
        default=None,
        help="The directory you want the report output to go. If not provided, it will overwrite "
        "the input report artifacts.",
    )
    parser.add_argument(
        "--template-filename",
        "-t",
        required=False,
        default=os.path.join(here, "templates", "report.html"),
        help="The name of a template to use when generating the report.",
    )
    return parser


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    args = get_parser().parse_args()
    output_dir = args.output_dir or os.path.dirname(args.input_json)
    exporter = ReportExporter(
        template_dir=os.path.dirname(args.template_filename), root_template=os.path.basename(args.template_filename)
    )
    report = Report.parse_file(args.input_json)
    exporter.export_all(report, output_dir)
    print(f"Exported artifacts to {output_dir}")



import base64
import os
import re
from datetime import datetime
from enum import Enum
from typing import List, Optional, Union

from pydantic import BaseModel, Field, validator


class Image(BaseModel):
    url: str
    base64: Optional[str]
    caption: Optional[str]
    is_error: bool = False
    source_url: Optional[str]

    def save(self, root: str):
        if not self.base64:
            return
        bytes_ = base64.b64decode(self.base64.encode("UTF-8"))
        filename = os.path.join(root, self.url)
        dirname = os.path.dirname(filename)
        os.makedirs(dirname, exist_ok=True)
        with open(filename, "wb") as f:
            f.write(bytes_)


class Outcome(Enum):
    success = "success"
    failure = "failure"
    never_started = "never started"


class Timed(BaseModel):
    start_time: datetime = Field(default_factory=lambda: datetime.now())
    end_time: Optional[datetime]

    def __enter__(self):
        self.start_time = datetime.now()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.end_time = datetime.now()

    def stop_timer(self):
        self.end_time = datetime.now()

    @property
    def duration(self) -> str:
        """
        Creates a human-readable minutes/seconds slug
        detailing how long the test took.

        If the duration was 129.5 seconds,
        the output would be '2m 10s'
        """
        end_time = self.end_time or datetime.now()
        minutes = 0
        seconds = (end_time - self.start_time).seconds
        if seconds > 60:
            minutes = int(seconds / 60)
            seconds = round(seconds % 60)
        if minutes:
            return f"{minutes}m {seconds}s"
        return f"{seconds}s"

    # There are several 'noqa: E821' notes below.
    # The arguments here are part of the dict() method provided by
    # pydantic, but refer to types we don't care about here. Rather
    # than unnecessarily mangle the signature of this complex
    # method, it's better to just ignore the E821 that complains
    # about it; it's "not our problem" because we didn't make
    # these signatures.
    def dict(
        self,
        *,
        include: Union["AbstractSetIntStr", "MappingIntStrAny"] = None,  # noqa: F821
        exclude: Union["AbstractSetIntStr", "MappingIntStrAny"] = None,  # noqa: F821
        by_alias: bool = False,
        skip_defaults: bool = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
    ) -> "DictStrAny":  # noqa: F821
        result = super().dict(
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            skip_defaults=skip_defaults,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
        )

        if not exclude or "duration" not in exclude:
            result["duration"] = self.duration

        return result


class TestResult(Timed):
    test_name: str
    test_id: Optional[str]  # Will be set from test_name
    pngs: List[Image] = []
    console_errors: List[str] = []
    traceback: Optional[str]
    test_description: Optional[str]
    outcome: Outcome = Outcome.never_started

    @validator("test_id", always=True)
    def populate_test_id(cls, v, values):
        test_id = re.sub(r"[^\w]", "-", values.get("test_name"))
        if test_id.endswith("-"):
            test_id = test_id[:-1]
        test_id = re.sub(r"--", "-", test_id)
        return test_id


class Report(Timed):
    outcome: Outcome = "never started"
    results: List[TestResult] = []
    arguments: Optional[str]
    title: str

    @property
    def failures(self) -> List[TestResult]:
        filter_ = filter(lambda result: (result.outcome != Outcome.success), self.results)
        return list(filter_)

    @property
    def num_failures(self) -> int:
        return len(self.failures)


class ReportResult:
    """
    A test result for passing to the report_test fixture.
    report -- a pytest test outcome
    excinfo -- exception info if there is any
    doc -- the docstring for the test if there is any
    """

    def __init__(self, report, excinfo, doc):
        self.report = report
        self.excinfo = excinfo
        self.doc = doc



import logging
import os
import shutil
from datetime import datetime
from typing import NoReturn

import jinja2

from webdriver_recorder.models import Outcome, Report

logger = logging.getLogger(__name__)
here = os.path.dirname(os.path.abspath(__file__))


class ReportExporter:
    def __init__(
        self,
        template_dir: str = os.path.join(here, "templates"),
        root_template: str = "report.html",
        static_assets_dir: str = "static",
    ):
        self.template_dir = template_dir
        self.relative_static_dir = static_assets_dir
        self.abs_input_static_dir = os.path.join(self.template_dir, static_assets_dir)
        self.static_assets = []
        if static_assets_dir:
            if os.path.exists(self.abs_input_static_dir):
                self.static_assets = [
                    os.path.join(self.abs_input_static_dir, asset) for asset in os.listdir(self.abs_input_static_dir)
                ]
            else:  # pragma: no cover
                logger.warning(
                    f"Expected {self.abs_input_static_dir} to exist, but it does not. Skipping static asset "
                    f"copying. You can disable this warning by setting `static_assets_dir=None` when creating "
                    f"your ReportExporter instance."
                )

        self.env = jinja2.Environment(loader=jinja2.FileSystemLoader(template_dir))
        self.env.filters["pretty_datetime"] = self.pretty_datetime
        self.template = self.env.get_template(root_template)

    @staticmethod
    def pretty_datetime(dt: datetime):
        return dt.strftime("%Y-%m-%d %H:%M:%S")

    def export_json(
        self, report: Report, dest_directory: str, dest_filename: str = "report.json", exclude_image_data: bool = False
    ) -> str:
        if report.num_failures > 0:
            report.outcome = Outcome.failure

        exclude = None
        if exclude_image_data:
            # This is a complex pydantic filter meaning:
            #   for each result in report,
            #   for each png in result,
            #   ignore the 'base64' field.
            logger.info("Stripping base64 image data from report.json")
            exclude = {"results": {"__all__": {"pngs": {"__all__": {"base64"}}}}}
        filename = os.path.join(dest_directory, dest_filename)
        with open(filename, "w") as f:
            f.write(report.json(indent=4, exclude=exclude))
        logger.info(f"Report JSON saved to {filename}")
        return filename

    def export_static(self, dest_directory: str):
        abs_dest_static_dir = os.path.join(dest_directory, self.relative_static_dir)
        for asset in self.static_assets:
            destination = os.path.abspath(os.path.join(abs_dest_static_dir, os.path.basename(asset)))
            os.makedirs(os.path.dirname(destination), exist_ok=True)
            shutil.copyfile(asset, destination)
            logger.info(f"Copied static asset {asset} to {destination}")

    def export_all(self, report: Report, dest_directory: str):
        # If we write images to disk, don't include base64 of the
        # images in the exported json.
        # This will still preserve image metadata, including its filename.

        self.export_json(report, dest_directory=dest_directory, exclude_image_data=True)
        self.export_images(report, dest_directory=dest_directory)
        self.export_static(dest_directory=dest_directory)
        self.export_html(report, dest_directory=dest_directory)

    def export_html(self, report: Report, dest_directory: str, dest_filename: str = "index.html"):
        dest_filename = os.path.join(dest_directory, dest_filename)
        stream = self.template.stream(report=report)
        stream.dump(dest_filename)
        logger.info(f"Exported report HTML to {dest_filename}")

    @classmethod
    def export_images(cls, report: Report, dest_directory: str) -> NoReturn:
        for test in report.results:
            for image in filter(lambda i: i.base64, test.pngs):
                image.save(dest_directory)
                logger.info(f"Saved image {image.url}")



import json
import pprint
import string
import time
from contextlib import contextmanager
from enum import Enum
from hashlib import sha256
from logging import getLogger
from typing import Any, Callable, List, Optional, Tuple, Union
from urllib.parse import urlparse

import selenium.webdriver.remote.webdriver
from pydantic import BaseModel, Field
from selenium import webdriver
from selenium.common.exceptions import WebDriverException
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.by import By as By_
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait

from webdriver_recorder.models import Image

logger = getLogger(__name__)

__all__ = [
    "BrowserRecorder",
    "Waiter",
    "BrowserError",
    "Chrome",
    "Remote",
    "Locator",
    "XPathWithSubstringLocator",
    "By",
]

_XPATH_TRANSLATE_CASE = f"translate(., '{string.ascii_uppercase}', '{string.ascii_lowercase}')"


class By(Enum):
    """
    An Enum based on selenium's By object, so that values can be explicitly declared.
    """

    ID = By_.ID
    XPATH = By_.XPATH
    LINK_TEXT = By_.LINK_TEXT
    PARTIAL_LINK_TEXT = By_.PARTIAL_LINK_TEXT
    NAME = By_.NAME
    TAG_NAME = By_.TAG_NAME
    CLASS_NAME = By_.CLASS_NAME
    CSS_SELECTOR = By_.CSS_SELECTOR


class Locator(BaseModel):
    """
    A base class for defining selenium search locators. Essentially this is meant to just be
    passed into selenium's _find_element/_find_elements functions. In some of the high-level selenium abstractions,
    this behavior is invisible, or must be late-bound. This allows for more flexibility with the webdriver_recorder
    API.

    Usage:

        danger_locator = Locator(search_method=SearchMethod.CSS_SELECTOR, search_value='div.panel.panel-danger')
        browser.find_element(*danger_locator.payload)
    """

    search_method: By
    search_value: Optional[str]

    @property
    def payload(self) -> Tuple:
        return self.search_method.value, self._search_value

    @property
    def _search_value(self) -> str:
        return self.search_value or ""

    @property
    def description(self) -> str:
        desc = f"{self.search_method.value}"
        if self.search_value:
            desc = f'{desc} whose value is "{self.search_value}"'
        return desc


class XPathWithSubstringLocator(Locator):
    """
    The CSS spec does not allow for selectors based on element text, making XPath ideal for
    such searches. This subclass searches for a given tag with the substring displayed; matches are
    case-insensitive.

    locator = XPathWithSubstringLocator(tag='div', displayed_substring='hello')  # will match <div>HELLO</div>
    """

    search_method = Field(By.XPATH, const=True)
    tag: str
    displayed_substring: str

    @property
    def _search_value(self) -> str:
        return _xpath_contains(f"//{self.tag}", self.displayed_substring)

    @property
    def description(self) -> str:
        desc = f"tag[{self.tag}]"
        if self.displayed_substring:
            desc = f'{desc} containing the string "{self.displayed_substring}"'
        return desc


class BrowserRecorder(selenium.webdriver.remote.webdriver.WebDriver):
    """
    A selenium webdriver with extra convenience utilities and
    automatic screenshot capturing.
    """

    pngs: List[Image] = []  # store screenshots here. intentionally global

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.maximize_window()
        self.autocapture = True  # automatically capture screenshots
        self.default_wait = kwargs.get("default_wait", 5)

    @contextmanager
    def autocapture_off(self):
        """Context manager temporarily disabling automatic screenshot generation."""
        previous_autocapture = self.autocapture  # for nesting
        self.autocapture = False
        try:
            yield
        finally:
            self.autocapture = previous_autocapture

    def _resolve_timeout(self, timeout_in: Optional[int]):
        if timeout_in is None:
            return self.default_wait
        return timeout_in

    def clear(self):
        """Clear the active element."""
        self.switch_to.active_element.clear()

    def click(self, locator: Locator, **kwargs):
        if "caption" not in kwargs:
            kwargs["caption"] = f"Click on {locator.description}"

        element = self.wait_until(locator, EC.element_to_be_clickable, **kwargs)
        element.click()
        return element

    def click_tag(self, tag: str, with_substring: str, **kwargs):
        """See wait_for_tag; this does the same thing for clicking on random elements."""
        return self.click(
            XPathWithSubstringLocator(tag=tag, displayed_substring=with_substring),
            **kwargs,
        )

    def find_element(self, by: Union[By, str] = By_.ID, value: Optional[Any] = None) -> WebElement:
        """Overrides the base find_element method to support the 'By' enum"""
        if isinstance(by, By):
            by = by.value
        return super().find_element(by, value)

    def find_elements(self, by: Union[By, str] = By_.ID, value: Optional[Any] = None) -> List[WebElement]:
        """Overrides the base find_element method to support the 'By' enum"""
        if isinstance(by, By):
            by = by.value
        return super().find_elements(by, value)

    def click_button(self, substring: str = "", **kwargs):
        """
        Wait for a button with substring to become clickable then click it.
        """
        return self.click_tag("button", substring, **kwargs)

    def wait_for(self, locator: Locator, **kwargs):
        """Wait for tag containing substring to show up in the DOM."""
        if "caption" not in kwargs:
            kwargs["caption"] = f"Wait for {locator.description}"

        return self.wait_until(locator, EC.visibility_of_element_located, **kwargs)

    def wait_until(
        self, locator: Locator, condition: Callable, timeout: Optional[int] = None, capture_delay: int = 0, **kwargs
    ):
        timeout = self._resolve_timeout(timeout)
        with self.wrap_exception(locator.description):
            wait = Waiter(self, timeout)
            return wait.until(
                condition(locator.payload),
                capture_delay=capture_delay,
                **kwargs,
            )

    def wait_for_tag(self, tag: str, with_substring: str, **kwargs):
        """
        Since the XPathWithSubstringLocator is still the best way to search by text, this shim provides
        easy access to it for dependents who want to upgrade to V3.0 but not have to use XPathWithSubstringLocator
        all over the place. This has the same signature as the legacy `wait_for`.
        """
        return self.wait_for(
            XPathWithSubstringLocator(tag=tag, displayed_substring=with_substring),
            **kwargs,
        )

    def run_commands(self, commands):
        """
        Run a series of commands. The first item in a command is the
        browser method to call and the remaining items are the args to
        pass it.
        """
        for method, *args in commands:
            getattr(self, method)(*args)

    @contextmanager
    def _resize_for_screenshot(self):
        original_size = self.get_window_size()
        required_width = self.execute_script("return document.body.parentNode.scrollWidth")
        required_height = self.execute_script("return document.body.parentNode.scrollHeight")
        self.set_window_size(required_width, required_height)
        yield
        self.set_window_size(original_size["width"], original_size["height"])

    def get(self, url: string, snap: bool = False, caption: Optional[str] = None):
        self.execute_script("console.clear();")
        super().get(url)
        if self.autocapture and snap:
            if not caption:
                caption = f"Render {url}"
            self.snap(caption=caption)

    @property
    def current_url_brief(self):
        url = urlparse(self.current_url)
        url = f"{url.netloc}{url.path}"
        return url

    def snap(self, caption: Optional[str] = None, is_error: bool = False):
        """
        Store the screenshot as a base64 png in memory.
        Resize the window ahead of time so the full page shows in the shot.
        """
        with self._resize_for_screenshot():
            # The 'body' tag is the preferred way of capturing a screenshot
            # because it will be more consistent than having to worry about
            # scrollbars, etc. However, in some cases, a body DOM
            # may have content but no calculated height, which causes an
            # error when trying to capture the screenshot.
            # This if/else prevents the error by capturing a screenshot
            # using the window if the body's height is false-y.
            body = self.find_element(By_.TAG_NAME, "body")
            if body.rect.get("height"):
                b64_image = body.screenshot_as_base64
            else:
                b64_image = self.get_screenshot_as_base64()
        # The sha256 digest is used to fingerprint the image.
        # This can SIGNIFICANTLY reduce the payload of a report
        # artifact bundle by de-duplicating images, especially
        # in parametrized tests.
        b64_sha = sha256(b64_image.encode("UTF-8")).hexdigest()
        self.pngs.append(
            Image(
                url=f"screenshots/{b64_sha}.png",
                base64=b64_image,
                caption=caption,
                is_error=is_error,
                source_url=self.current_url_brief,
            )
        )

    def send(self, *strings):
        """
        Send the list of strings to the window, with a TAB between each
        string.
        """
        chain = ActionChains(self)
        chain.send_keys(Keys.TAB.join(strings)).perform()

    def send_inputs(self, *strings):
        elements = self.find_elements(By.TAG_NAME.value, "input")
        for element, val in zip(elements, strings):
            element.send_keys(val)

    def hide_inputs(self):
        """Obscure all text inputs on the current screen."""
        javascript = """
            var inputsToHide = document.querySelectorAll('input');
            Array.prototype.forEach.call(inputsToHide, function(el){
                if (!el.attributes['type'] || el.attributes['type'].value === 'text')
                    el.setAttribute('type', 'password');
            })
        """
        self.execute_script(javascript)

    def open_tab(self):
        self.execute_script("window.open('');")
        self.switch_to.window(self.window_handles[-1])

    def close_tab(self):
        self.close()
        if self.window_handles:
            self.switch_to.window(self.window_handles[-1])

    @contextmanager
    def tab_context(self):
        self.open_tab()
        try:
            yield
        finally:
            self.close_tab()

    def send_secret(self, *encrypted_strings):
        """
        Send the list of strings to the window, decrypting them first.
        We do this encrypted to keep the secrets from appearing in
        Tracebacks.
        """
        strings = [self.decrypt(string) for string in encrypted_strings]
        self.send(*strings)

    def decrypt(self, encrypted_text):  # pragma: no cover
        """Method to decrypt text to be overridden."""
        raise NotImplementedError("decrypt should be overridden")

    @contextmanager
    def wrap_exception(self, message):
        """Wrap any exceptions caught in a BrowserError with message."""
        try:
            yield
        except Exception as e:
            if not isinstance(e, BrowserError):
                err = BrowserError(self, message)
                err.orig = e
            else:
                err = e

            # Only capture this screenshot if the error occurred
            # in a context that didn't automatically log the error.
            if not self.pngs or not self.pngs[-1].is_error:
                try:
                    self.snap(caption=f"Python error: {type(e).__name__}", is_error=True)
                except WebDriverException:  # pragma: no cover
                    logger.warning(f"Could not take screenshot after encountering error {e=}.")

            raise err from None


class Waiter(WebDriverWait):
    """Custom WebDriverWait object that grabs a screenshot after every wait."""

    def __init__(self, driver, timeout, *args, **kwargs):
        super().__init__(driver, timeout, *args, **kwargs)
        self.__driver = driver

    def until(
        self, *args, capture_delay: int = 0, caption: Optional[str] = None, is_error: Optional[bool] = False, **kwargs
    ) -> WebElement:
        """
        Every time we wait, take a screenshot of the outcome.
        capture_delay - when we're done waiting, wait just  a little longer
          for whatever animations to take effect.
        """
        found = False
        caption = caption or ""
        err = None
        try:
            element = super().until(*args, **kwargs)
            found = True
        except Exception as e:
            err = BrowserError(self.__driver, str(e))
            err.orig = e
            raise err from None
        finally:
            if self.__driver.autocapture or err:
                if capture_delay:
                    time.sleep(capture_delay)
                self.__driver.snap(caption=caption, is_error=not found)
        return element


class BrowserError(Exception):
    """Error to raise for a meaningful browser error report."""

    def __init__(self, browser: BrowserRecorder, message, *args):
        self.message = message
        self.url = browser.current_url
        self.logs = browser.get_log("browser")
        self.log_last_http(browser)
        self.orig = None
        super().__init__(message, self.url, self.logs, *args)

    @staticmethod
    def log_last_http(browser):
        """Log the last http transaction as an error."""

        if "har" not in browser.log_types:
            return
        logs = browser.get_log("har")
        if not logs:
            return
        last_message = json.loads(logs[-1].get("message", {}))
        entries = last_message.get("log", {}).get("entries", [])
        if not entries:
            return
        message = pprint.pformat(entries[-1])
        logger.error(f"Last HTTP transaction: {message}")


def _xpath_contains(node, substring):
    if '"' in substring:
        raise ValueError("double quotes in substring not supported")
    substring = substring.lower()
    return f'{node}[contains({_XPATH_TRANSLATE_CASE}, "{substring}")]'


class Chrome(BrowserRecorder, webdriver.Chrome):
    pass


class Remote(BrowserRecorder, webdriver.Remote):
    pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['webdriver_recorder']

package_data = \
{'': ['*'], 'webdriver_recorder': ['templates/*', 'templates/static/*']}

install_requires = \
['Jinja2>=3.0.1,<4.0.0',
 'pydantic>=1.8.2,<2.0.0',
 'pytest>=6.2.4,<7.0.0',
 'selenium>=4.1.0,<5.0.0']

entry_points = \
{'pytest11': ['uw-webdriver-recorder = webdriver_recorder.plugin']}

setup_kwargs = {
    'name': 'uw-webdriver-recorder',
    'version': '5.0.3',
    'description': 'A pytest plugin for recording screenshots of selenium interactions, with other convenient features too.',
    'long_description': None,
    'author': 'Tom Thorogood',
    'author_email': 'goodtom@uw.edu',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import logging
import os
import sys
import tempfile
from contextlib import contextmanager
from typing import Callable, Dict, List, Optional, Type, Union

import pytest
from pydantic import BaseSettings, validator
from selenium import webdriver

from .browser import BrowserError, BrowserRecorder, Chrome, Remote
from .models import Outcome, Report, ReportResult, TestResult, Timed
from .report_exporter import ReportExporter

_here = os.path.abspath(os.path.dirname(__file__))
logger = logging.getLogger(__name__)


class EnvSettings(BaseSettings):
    """
    Automatically derives from environment variables and
    translates truthy/falsey strings into bools. Only required
    for code that must be conditionally loaded; all others
    should be part of 'pytest_addoption()'
    """

    # If set to True, will generate a new browser instance within every request
    # for a given scope, instead of only creating a single instance and generating
    # contexts for each test.
    # This has a significant performance impact,
    # but sometimes cannot be avoided.
    disable_session_browser: Optional[bool] = False

    @validator("*", pre=True, always=True)
    def handle_empty_string(cls, v):
        if not v:
            return None
        return v


_SETTINGS = EnvSettings()


def pytest_addoption(parser):
    group = parser.getgroup("webdriver_recorder")
    group.addoption(
        "--selenium-server",
        action="store",
        dest="selenium_server",
        default=os.environ.get("REMOTE_SELENIUM"),
        help="Remote selenium webdriver to connect to (eg localhost:4444)",
    )
    group.addoption(
        "--report-dir",
        action="store",
        dest="report_dir",
        default=os.environ.get("REPORT_DIR", os.path.join(os.getcwd(), "webdriver-report")),
        help="The path to the directory where artifacts should be stored.",
    )
    group.addoption(
        "--jinja-template",
        action="store",
        dest="report_template",
        default=os.path.join(_here, "report.template.html"),
    )
    group.addoption(
        "--report-title",
        action="store",
        dest="report_title",
        default="Webdriver Recorder Summary",
        help="An optional title for your report; if not provided, a default will be used. "
        "You may also provide a constant default by overriding the report_title fixture.",
    )


@pytest.fixture(scope="session", autouse=True)
def clean_screenshots(report_dir):
    screenshots_dir = os.path.join(report_dir, "screenshots")
    if os.path.exists(screenshots_dir):
        old_screenshots = os.listdir(screenshots_dir)
        for png in old_screenshots:
            os.remove(os.path.join(screenshots_dir, png))


@pytest.fixture(scope="session", autouse=True)
def test_report(report_title) -> Report:
    args = []
    if len(sys.argv) > 1:
        args.extend(sys.argv[1:])

    return Report(
        arguments=" ".join(args),
        outcome=Outcome.never_started,
        title=report_title,
    )


@pytest.fixture(scope="session")
def selenium_server(request) -> Optional[str]:
    """Returns a non-empty string or None"""
    value = request.config.getoption("selenium_server")
    if value:
        return value.strip()
    return None


@pytest.fixture(scope="session")
def chrome_options() -> webdriver.ChromeOptions:
    """
    An extensible instance of ChromeOptions with default
    options configured for a balance between performance
    and test isolation.

    You can extend this:
        @pytest.fixture(scope='session')
        def chrome_options(chrome_options) -> ChromeOptions:
            chrome_options.add_argument("--option-name")
            return chrome_options

    or override it entirely:
        @pytest.fixture(scope='session')
        def chrome_options() -> ChromeOptions:
            return ChromeOptions()
    """
    options = webdriver.ChromeOptions()

    # Our default options promote a balance between
    # performance and test isolation.
    options.add_argument("--headless")
    options.add_argument("--incognito")
    options.add_argument("--disable-application-cache")
    options.add_argument("--no-sandbox")
    options.add_argument("--disable-dev-shm-usage")
    return options


@pytest.fixture(scope="session")
def browser_args(selenium_server, chrome_options) -> Dict[str, Optional[Union[webdriver.ChromeOptions, str]]]:
    args = {"options": chrome_options}
    if selenium_server:
        args["command_executor"] = f"http://{selenium_server}/wd/hub"
    return args


@pytest.fixture(scope="session")
def browser_class(browser_args) -> Type[BrowserRecorder]:
    if browser_args.get("command_executor"):
        return Remote
    return Chrome


@pytest.fixture(scope="session")
def build_browser(browser_args, browser_class) -> Callable[..., BrowserRecorder]:
    logger.info(
        "Browser generator will build instances using the following settings:\n"
        f"   Browser class: {browser_class.__name__}\n"
        f"   Browser args: {dict(browser_args)}"
    )

    def inner() -> BrowserRecorder:
        return browser_class(**browser_args)

    return inner


@pytest.fixture(scope="session")
def session_browser(build_browser) -> BrowserRecorder:
    """
    A browser instance that is kept open for the entire test run.
    Only instantiated if it is used, but by default will be used in both the
    'browser' and 'class_browser' fixtures, unless "disable_session_browser=1"
    is set in the environment.
    """
    browser = build_browser()
    try:
        yield browser
    finally:
        browser.quit()


@pytest.fixture(scope="session")
def browser_context() -> Callable[..., Chrome]:
    """
    This fixture allows you to create a fresh context for a given
    browser instance.

    The default behavior of the `browser` fixture is to always run in a context of the session scope, so
    you only need to use this if you are not using (or are overriding) the `browser` fixture.

    The fixture itself simply passes the context manager, so you can use it like so:

        def test_something(browser_context):
            with browser_context() as browser:
                browser.get('https://www.uw.edu')

    You may also provide a list of urls to visit to clear cookies at the end of your session,
    if the default 'delete_all_cookies' behavior is not enough to cover your use case.
    """

    @contextmanager
    def inner(browser: BrowserRecorder, cookie_urls: Optional[List[str]] = None) -> BrowserRecorder:
        browser.open_tab()
        cookie_urls = cookie_urls or []
        try:
            yield browser
        finally:
            browser.delete_all_cookies()
            for url in cookie_urls:
                browser.get(url)
                browser.delete_all_cookies()
            browser.close_tab()

    return inner


if _SETTINGS.disable_session_browser:
    logger.warning("Disabling auto-use of 'session_browser', this may significantly decrease test performance.")

    @pytest.fixture(scope="session")
    def session_browser_disabled() -> bool:
        return True

    @pytest.fixture
    def browser(build_browser) -> BrowserRecorder:
        """Creates a fresh instance of the browser using the configured chrome_options fixture."""
        browser = build_browser()
        try:
            yield browser
        finally:
            browser.quit()

    @pytest.fixture(scope="class")
    def class_browser(build_browser, request) -> BrowserRecorder:
        """
        Creates a fresh instance of the browser for use in the requesting class, using the configure
        chrome_options fixture.
        """
        browser = build_browser()
        request.cls.browser = browser
        try:
            yield browser
        finally:
            browser.quit()

else:
    logger.info(
        "Enabling auto-use of 'session_browser'; if your tests appear stuck, try disabling "
        "by setting 'disable_session_browser=1' in your environment."
    )

    @pytest.fixture
    def browser(session_browser, browser_context) -> BrowserRecorder:
        """
        Creates a function-scoped tab context for the session_browser which cleans
        up after itself (to the best of its ability). If you need a fresh instance
        each test, you can set `disable_session_browser=1` in your environment.
        """
        with browser_context(session_browser) as browser:
            yield browser

    @pytest.fixture(scope="class")
    def class_browser(request, session_browser, browser_context) -> BrowserRecorder:
        """
        Creates a class-scoped tab context and binds it to the requesting class
        as 'self.browser'; this tab will close once all tests in the class have run,
        and will clean up after itself (to the best of its ability). If you need
        a fresh browser instance for each class, you can set `disable_session_browser=1` in your
        environment.
        """
        with browser_context(session_browser) as browser:
            request.cls.browser = browser
            yield browser

    @pytest.fixture(scope="session")
    def session_browser_disabled() -> bool:
        return False


@pytest.fixture(scope="session")
def report_dir(request):
    dir_ = request.config.getoption("report_dir")
    os.makedirs(dir_, exist_ok=True)
    return dir_


@pytest.fixture(scope="session", autouse=True)
def report_generator(report_dir, test_report):
    with tempfile.NamedTemporaryFile(prefix="worker.", dir=report_dir) as worker_file:
        suffix = ".".join(worker_file.name.split(".")[1:])
        yield
    test_report.stop_timer()
    exporter = ReportExporter()
    workers = list(f for f in os.listdir(report_dir) if f.startswith("worker."))
    worker_results = list(f for f in os.listdir(report_dir) if f.endswith(".result.json"))
    if not workers:
        test_report.outcome = Outcome.success
        # Aggregate worker reports into this "root" report.
        for result_file in [os.path.join(report_dir, f) for f in worker_results]:
            worker_report = Report.parse_file(result_file)
            test_report.results.extend(worker_report.results)
            os.remove(result_file)
        exporter.export_all(test_report, report_dir)
    else:
        # If there are other workers, only export the report json of the
        # current worker. The last worker running will be responsible for aggregating and reporting results.
        exporter.export_json(test_report, report_dir, dest_filename=f"{suffix}.result.json")


@pytest.fixture(autouse=True)
def report_test(report_generator, request, test_report):
    """
    Print the results to report_file after a test run. Without this, the results of the test will not be saved.
    """
    tb = None
    console_logs = []
    timer: Timed
    with Timed() as timer:
        yield

    call_summary = getattr(request.node, "report_result", None)

    if call_summary:
        doc = call_summary.doc
        test_name = call_summary.report.nodeid
        outcome = Outcome.failure if call_summary.report.failed else Outcome.success
        if call_summary and call_summary.excinfo and not tb:
            outcome = Outcome.failure
            exception: BaseException = call_summary.excinfo.value
            exception_msg = f"{exception.__class__.__name__}: {str(exception)}"
            if isinstance(exception, BrowserError):
                if exception.orig:
                    tb = f"{exception_msg}\n{exception.orig=}"
                console_logs = [log.get("message", "") for log in exception.logs]
            if not tb:
                tb = f"{exception_msg}\n(No traceback is available)"

    else:
        logging.error(
            f"Test {request.node} reported no outcomes; "
            f"this usually indicates a fixture caused an error when setting up the test."
        )
        doc = None
        test_name = f"{request.node.name}"
        outcome = Outcome.never_started

    # TODO: Figure out a way to include class docs if they exist
    #       class TestFooBar:
    #           """
    #           When Foo is bar
    #           """
    #           def test_a_baz(self):
    #               """and baz is bop"""
    #               do_work('bop')
    #   The report output should then read "When foo is bar and baz is bop"

    result = TestResult(
        pngs=BrowserRecorder.pngs,
        test_name=test_name,
        test_description=doc,
        outcome=outcome,
        start_time=timer.start_time,
        end_time=timer.end_time,
        traceback=tb,
        console_errors=console_logs,
    )
    BrowserRecorder.pngs = []

    test_report.results.append(result)


@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_runtest_makereport(item, call):
    """
    This gives us hooks from which to report status post test-run.
    """
    outcome = yield

    report = outcome.get_result()
    if report.when == "call":
        doc = getattr(getattr(item, "function", None), "__doc__", None)
        item.report_result = ReportResult(report=report, excinfo=call.excinfo, doc=doc)


@pytest.fixture(scope="session")
def report_title(request) -> str:
    return request.config.getoption("report_title")
